/*	
	Charlie ?

	variation autour d'un rseau de Kohonen ;
	(UNE dimension) ; d'aprs "Rseaux neuronaux"
	de Jean-philippe Rennard.

	Version / 18 dcembre 2010.
*/

// ----------------------------------------------------------------------------------------------------------------

#include "ext.h"							
#include "ext_obex.h"
#include "ext_common.h"
#include "stdlib.h"
#include "math.h"	
#include "time.h"		

#define TAILLE_VECTEUR_MAX	256
#define	TAILLE_CARTE_MAX	100
#define	TAILLE_LCD_MAX		800
#define	TAILLE_LCD_MIN		100

#define DEF_VALEUR_MAX		127
#define DEF_PAS				1.
#define DEF_TAILLE_VECTEUR	4
#define DEF_INFLUENCE		25
#define DEF_MUTATION		40
#define DEF_TAILLE_CARTE	50
#define DEF_APPRENTISSAGES	60
#define DEF_LCD				150
	
// ----------------------------------------------------------------------------------------------------------------

typedef struct _charlie {
	t_object	ob ;	
	double		**carte ;
	long		taille_carte ;
	long		taille_vecteur ;
	long		valeur_max ;
	long		temps ;
	long		mutation ;
	long		influence ;
	long		apprentissages ;
	long		lcd ;
	long		lock ;
	double		pas ;
	t_atom		*resultat ;
	void		*m_qelem ;
	void		*m_clock ;
	void		*m_out1 ;
	void		*m_out2 ;
	} t_charlie ;
	
// ----------------------------------------------------------------------------------------------------------------	

void		*charlie_new		(t_symbol *s, long argc, t_atom *argv) ;
t_max_err	charlie_set			(t_charlie *x, t_object *attr, long argc, t_atom *argv) ;
t_max_err	charlie_get			(t_charlie *x, t_object *attr, long *argc, t_atom **argv) ;
void		charlie_free		(t_charlie *x) ;
void		charlie_assist		(t_charlie *x, void *b, long m, long a, char *s) ;

void		charlie_clear		(t_charlie *x) ;
void		charlie_learn		(t_charlie *x, t_symbol *s, long argc, t_atom *argv) ;

void		charlie_list		(t_charlie *x, t_symbol *s, long argc, t_atom *argv) ;
void		charlie_int			(t_charlie *x, long n) ;
void		charlie_bang		(t_charlie *x) ;

void		carte_init			(t_charlie *x) ;

long		carte_gagnant		(t_charlie *x, double *q) ;
double		carte_influence		(t_charlie *x, long g, long n) ;
void		vecteur_modifier	(t_charlie *x, double *p, double *q, double phi) ;

void		charlie_dumpmax		(t_charlie *x) ;
void		charlie_task_max	(t_charlie *x) ;
void		charlie_dumplcd		(t_charlie *x) ;
void		charlie_qtask_lcd	(t_charlie *x) ;

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

static t_class *charlie_class ;

int main (void)
	{	
		t_class *c = NULL ;
		
		c = class_new ("charlie", (method)charlie_new, (method)charlie_free, (long)sizeof(t_charlie), 0L, A_GIMME, 0) ;

		class_addmethod (c, (method)charlie_assist,		"assist", A_CANT, 0) ; 
		class_addmethod (c, (method)charlie_bang,		"bang", 0) ;
		class_addmethod (c, (method)charlie_clear,		"clear", 0) ;
		class_addmethod (c, (method)charlie_dumpmax,	"dump_max", 0) ;
		class_addmethod (c, (method)charlie_dumplcd,	"dump_lcd", 0) ;
		class_addmethod (c, (method)charlie_learn,		"learn", A_GIMME, 0) ; 
		class_addmethod (c, (method)charlie_list,		"list", A_GIMME, 0) ;
		class_addmethod (c, (method)charlie_list,		"change", A_GIMME, 0) ;
		class_addmethod (c, (method)charlie_int,		"int", A_LONG, 0) ;
		
		class_register (CLASS_BOX, c) ; 
		
		CLASS_ATTR_LONG			(c, "lcd",			ATTR_FLAGS_NONE, t_charlie, lcd) ;
		CLASS_ATTR_LONG			(c, "mutation",		ATTR_FLAGS_NONE, t_charlie, mutation) ;	
		CLASS_ATTR_LONG			(c, "influence",	ATTR_FLAGS_NONE, t_charlie, influence) ;
		CLASS_ATTR_FLOAT		(c, "step",			ATTR_FLAGS_NONE, t_charlie, pas) ;
		
		CLASS_ATTR_LABEL		(c, "mutation",		ATTR_FLAGS_NONE, "Mutation") ;
		CLASS_ATTR_LABEL		(c, "lcd",			ATTR_FLAGS_NONE, "Lcd Size") ;
		CLASS_ATTR_LABEL		(c, "influence",	ATTR_FLAGS_NONE, "Influence Range") ;
		CLASS_ATTR_LABEL		(c, "step",			ATTR_FLAGS_NONE, "Learning Step") ;
		
		CLASS_ATTR_FILTER_CLIP	(c, "lcd",			TAILLE_LCD_MIN, TAILLE_LCD_MAX) ;
		CLASS_ATTR_FILTER_CLIP	(c, "mutation",		0, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "influence",	1, TAILLE_CARTE_MAX) ;
		CLASS_ATTR_FILTER_CLIP	(c, "step",			0., 1.) ;
		
		CLASS_ATTR_ORDER		(c, "mutation",		ATTR_FLAGS_NONE, "1") ;
		CLASS_ATTR_ORDER		(c, "lcd",			ATTR_FLAGS_NONE, "2") ;
		CLASS_ATTR_ORDER		(c, "influence",	ATTR_FLAGS_NONE, "3") ;
		CLASS_ATTR_ORDER		(c, "step",			ATTR_FLAGS_NONE, "4") ;
		
		CLASS_ATTR_ACCESSORS	(c, "step",			charlie_get, charlie_set) ;
				
		charlie_class = c ;
		
		return MAX_ERR_NONE ;
	}

// ----------------------------------------------------------------------------------------------------------------

void *charlie_new (t_symbol *s, long argc, t_atom *argv)
	{
		t_charlie	*x = NULL ;
		
		if (x = (t_charlie *)object_alloc (charlie_class)) 
			{
				long ac = 0 ;
				char alloc ;
				
				srand (time(NULL)) ;
					
				x->taille_vecteur	= DEF_TAILLE_VECTEUR ;
				x->mutation			= DEF_MUTATION ;
				x->pas				= DEF_PAS ;
				x->influence		= DEF_INFLUENCE ;
				x->apprentissages	= DEF_APPRENTISSAGES ;
				x->lcd				= DEF_LCD ;
				x->taille_carte		= DEF_TAILLE_CARTE ;
				x->valeur_max		= DEF_VALEUR_MAX ;
				x->lock				= 0 ;
				x->carte			= NULL ;
				x->resultat			= NULL ; 
				
				atom_alloc_array	(TAILLE_VECTEUR_MAX, &ac, &x->resultat, &alloc) ;
				
				if (argc)
					{
						if (atom_gettype (argv) == A_LONG)
							{
								long k = atom_getlong (argv) ;
								
								if (k > 1 && k <= TAILLE_CARTE_MAX)
									{
										x->taille_carte = k ;
										x->influence = k / 2 ;
									}
								
								if (argc > 1)
									{
										if (atom_gettype (argv + 1) == A_LONG)
											{
												long w = atom_getlong (argv + 1) ;
												
												if (w > 0 && w <= 1000)
													{
														x->apprentissages = w ;
													}
											}
										
										if (argc > 2)
											{
												if (atom_gettype (argv + 2) == A_LONG)
													{
														long z = atom_getlong (argv + 2) ;
														
														if (z > 1 && z <= LONG_MAX)
															{
																x->valeur_max = z ;
															}
													}
											}
									}
							}
					}
												
				attr_args_process (x, argc, argv) ;
								
				if (x->carte = (double **)sysmem_newptr (sizeof(double *) * x->taille_carte))
					{
						long j ;
				
						for (j = 0 ; j < x->taille_carte ; j++)
							{
								x->carte[j] = (double *)sysmem_newptrclear (sizeof(double) * TAILLE_VECTEUR_MAX) ; 
							}
					}
					
				carte_init (x) ;
				
				x->m_qelem  = qelem_new ((t_object *)x, (method)charlie_qtask_lcd) ;
				x->m_clock	= clock_new ((t_object *)x, (method)charlie_task_max) ;
				
				x->m_out2	= outlet_new ((t_object *)x, NULL) ;
				x->m_out1	= listout ((t_object *)x) ;
				
				if (ac != TAILLE_VECTEUR_MAX)
					{
						object_free (x) ;
						x = NULL ;
					}
			}
				
		return x ;
	}

t_max_err charlie_set (t_charlie *x, t_object *attr, long argc, t_atom *argv)
    {
		if (argc && argv)
			{
				x->pas = atom_getfloat (argv) ;
			}

        return MAX_ERR_NONE ;
    }
	
t_max_err charlie_get (t_charlie *x, t_object *attr, long *argc, t_atom **argv)
    {
        char	alloc ;
		double	k = x->pas ; 
		
		if (!((*argc) && (*argv))) 
			{
				if (atom_alloc (argc, argv, &alloc))  
					{
						return MAX_ERR_OUT_OF_MEM ;
					}
			}
			
        atom_setfloat (*argv, k) ;
		
        return MAX_ERR_NONE ;
    }
	
void charlie_free (t_charlie *x)
	{ 
		long i ;
		
		if (x->resultat)
			{
				sysmem_freeptr (x->resultat) ;
			}
			
		for (i = 0 ; i < x->taille_carte ; i++)
			{
				if (x->carte[i]) ;
					{
						sysmem_freeptr (x->carte[i]) ;
					}
			}
		
		if (x->carte)
			{
				sysmem_freeptr (x->carte) ;
			}
		
		if (x->m_qelem)
			{
				qelem_unset (x->m_qelem) ;
				qelem_free	(x->m_qelem) ;
			}
		
		if (x->m_clock)
			{
				clock_unset (x->m_clock) ;
				object_free (x->m_clock) ;
			}
	}

void charlie_assist (t_charlie *x, void *b, long m, long a, char *s)
	{
		if (m == ASSIST_INLET) 
			{ 
				sprintf (s, "(list) (int) (bang) clear learn change") ;
			} 
		else 
			{	
				if (a == 0 )
					{
						sprintf (s, "Values") ;	
					}
				else
					{
						sprintf (s, "Dump to lcd") ;
					}
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void charlie_clear (t_charlie *x) 
	{
		carte_init (x) ;
	}
	
void charlie_learn (t_charlie *x, t_symbol *s, long argc, t_atom *argv)
	{
		if (argc != x->taille_vecteur && argc <= TAILLE_VECTEUR_MAX)  
			{
				x->taille_vecteur = argc ;				
				carte_init (x) ;
			}
			
		if (argc == x->taille_vecteur)
			{
				long	i, g ;
				double	temp[TAILLE_VECTEUR_MAX] ;
				
				for (i = 0 ; i < argc ; i ++)
					{
						if (atom_gettype (argv + i) != A_LONG)
							{
								temp[i] = rand () % (x->valeur_max + 1) ;
							}
						else
							{
								temp[i] = CLIP ((double)atom_getlong (argv + i), 0., (double)x->valeur_max) ;
							}
					}
				
				g = carte_gagnant (x, temp) ;
				
				if (g >= 0)
					{
						long j ;
						
						if (x->temps < x->apprentissages)
							{
								x->temps ++ ;
							}
							
						for (j = 0 ; j < x->taille_carte ; j++) 
							{
								if (x->carte[j])
									{
										vecteur_modifier (x, x->carte[j], temp, carte_influence (x, g, j)) ;
									}
							}
					}
			} 
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
		
void charlie_list (t_charlie *x, t_symbol *s, long argc, t_atom *argv)
	{
		long	k = rand () % x->taille_carte ;
		t_atom	*aptr = x->resultat ;
		double	*ptr = x->carte[k] ;
		
		if (x->temps > 0 && ptr && aptr)
			{
				long i ;
				
				for (i = 0 ; i < MIN (argc, TAILLE_VECTEUR_MAX) ; i++)
					{
						long h = rand () % 100 ;
						
						if (h < x->mutation)
							{
								atom_setlong (aptr + i, (long)round (ptr[i % x->taille_vecteur])) ;
							}
						else
							{
								if (atom_gettype (argv + i) == 1)
									{
										atom_setlong (aptr + i, atom_getlong (argv + i)) ;
									}
								else
									{
										atom_setlong (aptr + i, 0) ;
									}
							}
					}
			}
		else
			{
				long i ;
				
				for (i = 0 ; i < MIN (argc, TAILLE_VECTEUR_MAX) ; i++)
					{
						if (atom_gettype (argv + i) == 1)
							{
								atom_setlong (aptr + i, atom_getlong (argv + i)) ;
							}
						else
							{
								atom_setlong (aptr + i, 0) ;
							}
					}
			}
			
		outlet_list (x->m_out1, NULL, (short)argc, x->resultat) ;
	}

void charlie_int (t_charlie *x, long n)
	{
		if (n >= 1 && n <= x->taille_carte)
			{
				double	*ptr = x->carte[n - 1] ;
				t_atom  *aptr = x->resultat ;
				
				if (ptr && aptr)
					{
						long i ;
						
						for (i = 0 ; i < x->taille_vecteur ; i++)
							{
								atom_setlong (aptr + i, (long)round (ptr[i])) ;
							}
							
						outlet_list (x->m_out1, NULL, (short)x->taille_vecteur, aptr) ;
					}	
			}
	}

void charlie_bang (t_charlie *x)
	{
		long	h = rand () % x->taille_carte ;
		double	*ptr = x->carte[h] ;
		t_atom  *aptr = x->resultat ;
				
		if (ptr && aptr)
			{
				long i ;
				
				for (i = 0 ; i < x->taille_vecteur ; i++)
					{
						atom_setlong (aptr + i, (long)round (ptr[i])) ;
					}
					
				outlet_list (x->m_out1, NULL, (short)x->taille_vecteur, aptr) ;
			}
	}
		
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void carte_init (t_charlie *x)
	{
		if (x->carte)
			{
				long i ;
		
				for (i = 0 ; i < x->taille_carte ; i++)
					{
						double *ptr = NULL ;
						
						if (ptr = x->carte[i])  
							{
								long j ;

								for (j = 0 ; j < x->taille_vecteur ; j++)
									{
										ptr[j] = (double)(rand () % (x->valeur_max + 1)) ;
									}
							}
					}
			}
		
		x->temps = 0 ;
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

long carte_gagnant (t_charlie *x, double *q)
	{
		long	i ;
		long	k = -1 ;
		double	a = 0. ;
		double	*p = NULL ;
		
		for (i = 0 ; i < x->taille_carte ; i++)
			{
				if (p = x->carte[i])
					{
						if (!i)
							{
								long j ;
								
								for (j = 0 ; j < x->taille_vecteur ; j++)
									{
										a += pow (p[j] - q[j], 2) ;
									}
								
								a = sqrt (a) ;
								k = i ;
							}
						else
							{
								long	j ;
								double	b = 0. ;
								
								for (j = 0 ; j < x->taille_vecteur ; j++)
									{
										b += pow (p[j] - q[j], 2) ;
									}
								
								b = sqrt (b) ;
				
								if (b < a)
									{
										a = b ;
										k = i ;
									}
							}
					}
			}
		
		return k ;
	}
	
double carte_influence (t_charlie *x, long g, long n) 
	{
		double tau, sigma, phi ;
		
		tau = x->apprentissages / log (x->influence) ;
		sigma = x->influence * exp (-(x->temps) / tau) ;
		phi = exp (-(pow (n - g, 2) / (2 * pow (sigma, 2)))) ;
		
		return phi ;
	}

void vecteur_modifier (t_charlie *x, double *p, double *q, double phi)
	{
		long	i ;
		double	eta = x->pas * exp (-(x->temps) / x->apprentissages) ;
		
		for (i = 0 ; i < x->taille_vecteur ; i++)
			{
				double k = p[i] ;
				double w = q[i] ;

				p[i] = k + (eta * phi * (w - k)) ;
			}
	}		

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void charlie_dumpmax (t_charlie *x) 
	{
		if (!x->lock)
			{
				long text_max = 100 ;
				char *affichage = NULL ;
				
				x->lock = 1 ;
				
				post ("Apprentissages : %ld / %ld", x->temps, x->apprentissages) ;
				post ("---") ;
				
				if (affichage = (char *)sysmem_newptr (sizeof(char) * text_max))
					{
						long i ;
						
						for (i = 0 ; i < x->taille_carte ; i++)
							{
								double *ptr = NULL ;
								
								if (ptr = x->carte[i])
									{
										long	j ;
										long	ac = x->taille_vecteur ;
										t_atom	temp[TAILLE_VECTEUR_MAX] ;
										
										for (j = 0 ; j < ac ; j++) 
											{
												atom_setlong (temp + j, (long)round (ptr[j])) ;
											}
										
										atom_gettext (ac, temp, &text_max, &affichage, OBEX_UTIL_ATOM_GETTEXT_DEFAULT) ;
										
										post ("%ld : %s", i + 1, affichage) ;
									}
							}
							
						sysmem_freeptr (affichage) ;
					}	
				
				post (" ") ;
				
				clock_fdelay (x->m_clock, 1000.) ;
			}
	}
	
void charlie_task_max (t_charlie *x)
	{
		x->lock = 0 ;
	}

void charlie_dumplcd (t_charlie *x) 
	{	
		qelem_set (x->m_qelem) ;
	}

void charlie_qtask_lcd (t_charlie *x)
    {
        long	i ;
		long	alpha = MAX (x->lcd / x->taille_carte, 2) ;
		long	beta  = MAX (x->lcd / x->taille_vecteur, 2) ;
		t_atom	argv[5] ;
		t_atom	lcdsize[2] ;
		
		atom_setlong (lcdsize, alpha * x->taille_carte) ;
		atom_setlong (lcdsize + 1, beta * x->taille_vecteur) ;
		
		outlet_anything (x->m_out2, gensym("clear"), 0, NULL) ;
		outlet_anything (x->m_out2, gensym("size"), 2, lcdsize) ;
		
		for (i = 0 ; i < x->taille_carte ; i++)
			{
				double *ptr = NULL ;
				
				if (ptr = x->carte[i])
					{
						long j ;
						
						for (j = 0 ; j < x->taille_vecteur ; j++)
							{
								long k = (long)round (((ptr[j]) / x->valeur_max) * 128) ;
								
								atom_setlong (&argv[0], i * alpha) ;
								atom_setlong (&argv[1], j * beta) ;
								atom_setlong (&argv[2], ((i + 1) * alpha)) ;
								atom_setlong (&argv[3], ((j + 1) * beta)) ;
								atom_setlong (&argv[4], CLIP (k, 0, 255)) ;
								
								outlet_anything (x->m_out2, gensym ("paintrect"), 5, argv) ;
							}
					}
			}
    }
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------:x