/*		

Romeo ! 
							  
(Treillis de Galois / Diagramme de Hasse).

Variation autour de : "Incremental concept formation algorithms based 
on Galois (concept) lattices", par Robert Godin, Rokia Missaoui, Hassan Alaoui. 
						
Version / 12 dcembre 2010.	

*/

// -------------------------------------------------------------------------------------------------------------

#include "ext.h"							
#include "ext_obex.h"
#include "stdlib.h"
#include "math.h"	
#include "time.h"	
#include "ext_common.h"
#include "ext_systhread.h"	

// -------------------------------------------------------------------------------------------------------------

#define TAILLE_VRAC				172
#define TAILLE_FIFO				170
#define MAX_CONCEPT				150

#define TAILLE_DESCRIPTEURS		128
#define TAILLE_ULONG			32
#define TAILLE_BLOCS			4

#define DEF_GRAF_MAX			52
#define DEF_MUTATION			50

/*
	TAILLE_FIFO = TAILLE_VRAC - 2 ;
	MAX_CONCEPT < TAILLE_FIFO ;
	
	TAILLE_DESCRIPTEURS = TAILLE_ULONG * TAILLE_BLOCS ;
*/

// -------------------------------------------------------------------------------------------------------------

typedef struct _concept {
	ulong			descripteurs[TAILLE_BLOCS] ;
	uchar			parents[TAILLE_VRAC] ;
	uchar			enfants[TAILLE_VRAC] ;
	long			cardinal ;
	} t_concept ;
		
// -------------------------------------------------------------------------------------------------------------

typedef struct _romeo {
	t_object			ob ;
	t_concept			*graf_vrac ;
	long				graf_max ;
	long				graf_pos ;
	long				graf_temp ;
	long				cible ;
	long				*bins ;
	long				*temp ;
	long				*bins_topo ;
	long				*temp_topo ;
	ulong				vecteur[TAILLE_BLOCS] ;
	ulong				intersection[TAILLE_BLOCS] ;
	long				cardinal_intersection ;
	long				maj ;
	long				nombre_concept ;
	long				mutation ;
	t_linklist			*resultat ;
	long				*fifo ;
	long				head_fifo ;
	long				queue_fifo ;
	long				val_fifo ;
	long				lock_dump ;
	t_systhread_mutex   m_mutex ;
	void				*m_clock ;
	void				*m_out1 ;
	void				*m_out2 ;
	} t_romeo ;
	
// -------------------------------------------------------------------------------------------------------------	
	
void	*romeo_new				(t_symbol *s, long argc, t_atom *argv) ;
long	romeo_init				(t_romeo *x) ;
void	romeo_free				(t_romeo *x) ;
void	romeo_assist			(t_romeo *x, void *b, long m, long a, char *s) ;

void	romeo_bang				(t_romeo *x) ;
void	romeo_random			(t_romeo *x) ;
void	romeo_clear				(t_romeo *x) ;
void	romeo_int				(t_romeo *x, long n) ;
void	romeo_learn				(t_romeo *x, t_symbol *s, long argc, t_atom *argv) ;
void    romeo_list				(t_romeo *x, t_symbol *s, long argc, t_atom *argv) ;

void	outlet_linkfun			(t_object *p, t_romeo *x) ;

void	graf_effacer			(t_romeo *x) ;
void	graf_ajouter			(t_romeo *x) ;
void	graf_cartographier		(t_romeo *x) ;
void	graf_tuer				(t_romeo *x) ;
long	graf_naviguer			(t_romeo *x) ;
long	graf_hasarder			(t_romeo *x) ;
long	graf_lire				(t_romeo *x, long n) ;

long	nouveau_concept			(t_romeo *x) ;
void	couverture_concept		(t_romeo *x, long h, long n) ;
void	tuer_concept			(t_romeo *x, long n) ;

long	sans_parents			(t_romeo *x, long a) ;
long	sans_enfants			(t_romeo *x, long b) ;

long	concept_inclusion		(t_romeo *x, ulong *ptr) ;
long	intersection_identique	(t_romeo *x, ulong *ptr) ;
long	intersection_inclusion	(t_romeo *x, ulong *ptr) ;

void	set_itemset				(ulong *ptr, long n) ;
long	get_itemset				(ulong *ptr, long n) ;
long	size_itemset			(ulong *ptr) ;

long	put_fifo				(t_romeo *x, long n) ;
long	get_fifo				(t_romeo *x) ;
void	clear_fifo				(t_romeo *x) ;

void	romeo_dump				(t_romeo *x, long n) ;
void	romeo_task				(t_romeo *x) ;

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

static t_class *romeo_class ;

int main (void)
	{	
		t_class *c = NULL ;
		
		c = class_new ("romeo", (method)romeo_new, (method)romeo_free, (long)sizeof(t_romeo), 0L, A_GIMME, 0) ;
		
		class_addmethod (c, (method)romeo_assist,	"assist", A_CANT, 0) ;
		class_addmethod (c, (method)romeo_dump,		"dump_max", A_DEFLONG, 0) ;
		class_addmethod (c, (method)romeo_bang,		"bang", 0) ;
		class_addmethod (c, (method)romeo_random,	"random", 0) ;
		class_addmethod (c, (method)romeo_int,		"int", A_LONG, 0) ;
		class_addmethod (c, (method)romeo_clear,	"clear", 0) ;
		class_addmethod (c, (method)romeo_learn,	"learn", A_GIMME, 0) ;
		class_addmethod (c, (method)romeo_list,		"list", A_GIMME, 0) ;
		class_addmethod (c, (method)romeo_list,		"change", A_GIMME, 0) ;
				
		class_register (CLASS_BOX, c) ; 
		
		CLASS_ATTR_LONG			(c, "mutation",	ATTR_FLAGS_NONE, t_romeo, mutation) ;
		CLASS_ATTR_LABEL		(c, "mutation",	ATTR_FLAGS_NONE, "Mutation") ;
		CLASS_ATTR_FILTER_CLIP	(c, "mutation",	0, 100) ;
		
		romeo_class = c ;
		
		return 0 ;
	}

void *romeo_new (t_symbol *s, long argc, t_atom *argv)
	{
		t_romeo *x = NULL;
		
		if (x = (t_romeo *)object_alloc (romeo_class)) 
			{	
				if (x->graf_vrac = (t_concept *)sysmem_newptrclear (sizeof(t_concept) * TAILLE_VRAC))
					{	
						long i ;
						
						srand (time(NULL)) ;
							
						x->lock_dump				= 0 ;
						x->cardinal_intersection	= 0 ;
						x->maj						= 0 ;
						x->nombre_concept			= 0 ;
						x->head_fifo				= 0 ;
						x->queue_fifo				= 0 ;
						x->val_fifo					= 0 ;
						x->cible					= -1 ;
						x->graf_pos					= -1 ;
						x->graf_temp				= -1 ;
						x->mutation					= DEF_MUTATION ;
						x->graf_max					= DEF_GRAF_MAX ;
						x->bins						= NULL ;
						x->temp						= NULL ;
						x->bins_topo				= NULL ;
						x->temp_topo				= NULL ;
						x->resultat					= NULL ;
						
						for (i = 0 ; i < TAILLE_BLOCS ; i++)
							{
								x->vecteur[i]		= 0 ;
								x->intersection[i]	= 0 ;
							}
						
						if (atom_gettype (argv) == A_LONG)
							{
								long k = atom_getlong (argv) ;
								
								if (k > 1 && k <= MAX_CONCEPT)
									{
										x->graf_max = k + 2 ;
									}
							}
						
						x->m_out2 = bangout ((t_object *)x) ;
						x->m_out1 = listout ((t_object *)x) ;
						
						x->m_clock	= clock_new ((t_object *)x, (method)romeo_task) ;
						
						systhread_mutex_new (&x->m_mutex, SYSTHREAD_MUTEX_NORMAL) ;
						
						attr_args_process (x, argc, argv) ;
								
						if (!romeo_init (x))
							{
								object_free (x) ;
								x = NULL ;
							}
					}
				else
					{
						object_free (x) ;
						x = NULL ;
					}
			}
				
		return x ;
	}

long romeo_init (t_romeo *x)
	{
		long i ;
		
		for (i = 0 ; i < TAILLE_DESCRIPTEURS ; i++)
			{
				set_itemset (x->graf_vrac[1].descripteurs, i) ;
			}
				
		x->graf_vrac[0].cardinal = 0 ;
		x->graf_vrac[1].cardinal = TAILLE_DESCRIPTEURS ;
		
		x->graf_vrac[0].enfants[1] = 1 ;
		x->graf_vrac[1].parents[0] = 1 ;
					
		if (x->bins_topo = (long *)sysmem_newptrclear (sizeof(long) * (TAILLE_DESCRIPTEURS + 1)))
			{
				x->bins_topo[0] ++ ;
				x->bins_topo[TAILLE_DESCRIPTEURS] ++ ;
			}
		
		if (x->bins = (long *)sysmem_newptrclear (sizeof(long) * ((TAILLE_DESCRIPTEURS + 1) * TAILLE_VRAC)))
			{
				x->bins[TAILLE_DESCRIPTEURS * TAILLE_VRAC] = 1 ;
			}
		
		x->temp			= (long *)sysmem_newptrclear (sizeof(long) * ((TAILLE_DESCRIPTEURS + 1) * TAILLE_VRAC)) ;
		x->temp_topo	= (long *)sysmem_newptrclear (sizeof(long) * (TAILLE_DESCRIPTEURS + 1)) ;
		
		if (x->resultat	= linklist_new ( )) 
			{
				linklist_flags (x->resultat, OBJ_FLAG_OBJ) ;
			}
		
		if (x->fifo	= (long *)sysmem_newptrclear (sizeof(long) * (TAILLE_FIFO + 1)))
			{
				long j ;
				
				for (j = 2 ; j < TAILLE_VRAC ; j++)
					{
						put_fifo (x, j) ;
					}
			}
		
		if (x->temp && x->temp_topo && x->bins && x->bins_topo && x->resultat && x->fifo)
			{
				return 1 ;
			}
		else
			{	
				return 0 ;
			}
	}

void romeo_free (t_romeo *x)
	{ 
		if (x->graf_vrac)
			{
				sysmem_freeptr (x->graf_vrac) ;
			}
		
		if (x->bins)
			{
				sysmem_freeptr (x->bins) ;
			}
		
		if (x->temp)
			{
				sysmem_freeptr (x->temp) ;
			}
		
		if (x->bins_topo)
			{
				sysmem_freeptr (x->bins_topo) ;
			}
		
		if (x->temp_topo)
			{
				sysmem_freeptr (x->temp_topo) ;
			}
		
		if (x->fifo)
			{
				sysmem_freeptr (x->fifo) ;
			}
			
		if (x->resultat)
			{
				object_free (x->resultat) ;
			}
		
		if (x->m_clock)
			{
				clock_unset (x->m_clock) ;
				object_free (x->m_clock) ;
			}
		
		if (x->m_mutex)
			{
				systhread_mutex_free (x->m_mutex) ;
			}
	}

void romeo_assist (t_romeo *x, void *b, long m, long a, char *s)
	{
		if (m == ASSIST_INLET) 
			{ 
				sprintf (s, "(list) (int) (bang) clear learn change random") ;
			} 
		else 
			{	
				switch (a) {
					case 0 :
						sprintf (s, "Navigate / get concepts") ;
						break ;
					case 1:
						sprintf (s, "Bang if empty") ;
						break ;
				}
			}
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void romeo_bang (t_romeo *x)
	{
		long k = 0 ;
		
		systhread_mutex_lock (x->m_mutex) ;

		if (x->nombre_concept != 0)
			{
				do {
				k = graf_naviguer (x) ;
				} while (!k) ;
			}
		
		systhread_mutex_unlock (x->m_mutex) ;
		
		if (!k)
			{
				outlet_bang (x->m_out2) ;
			}
		else
			{
				linklist_funall	(x->resultat, (method)outlet_linkfun, x) ;
				linklist_clear	(x->resultat) ;
			}
	}

void romeo_random (t_romeo *x)
	{
		long k = 0 ;
		
		systhread_mutex_lock (x->m_mutex) ;

		if (x->nombre_concept != 0)
			{
				k = graf_hasarder (x) ;
			}
		
		systhread_mutex_unlock (x->m_mutex) ;
		
		if (!k)
			{
				outlet_bang (x->m_out2) ;
			}
		else
			{
				linklist_funall	(x->resultat, (method)outlet_linkfun, x) ;
				linklist_clear	(x->resultat) ;
			}
	}

void romeo_clear (t_romeo *x)
	{
		systhread_mutex_lock (x->m_mutex) ;
		
		graf_effacer (x) ;
				
		systhread_mutex_unlock (x->m_mutex) ;
	}

void romeo_int (t_romeo *x, long n)
	{
		long k = 0 ;
		
		systhread_mutex_lock (x->m_mutex) ;

		if (x->nombre_concept != 0)
			{
				k = graf_lire (x, CLIP (n, 1, TAILLE_DESCRIPTEURS - 1)) ;
			}
		
		systhread_mutex_unlock (x->m_mutex) ;
		
		if (!k)
			{
				outlet_bang (x->m_out2) ;
			}
		else
			{
				linklist_funall	(x->resultat, (method)outlet_linkfun, x) ;
				linklist_clear	(x->resultat) ;
			}
	}

void romeo_learn (t_romeo *x, t_symbol *s, long argc, t_atom *argv)
	{
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{
				long i ;
				
				for (i = 0 ; i < TAILLE_BLOCS ; i++)
					{
						x->vecteur[i] = 0 ;
					}
				
				for (i = 0 ; i < argc ; i++)
					{
						if (atom_gettype (argv + i) == A_LONG)
							{
								long k = CLIP (atom_getlong (argv + i), 0, TAILLE_DESCRIPTEURS - 1) ;
																
								set_itemset (x->vecteur, k) ;
							}
					}
			
				graf_ajouter (x) ;
				
				systhread_mutex_unlock (x->m_mutex) ;
			}
	}

void romeo_list (t_romeo *x, t_symbol *s, long argc, t_atom *argv) 
	{
		long		k = 0 ;
		t_atomarray *temp = NULL ;
		
		systhread_mutex_lock (x->m_mutex) ;
		
		temp = atomarray_new (0, NULL) ;
		
		if (x->nombre_concept != 0)
			{
				do {
				k += graf_naviguer (x) ;
				} while (k < argc) ;
			}
		
		if (k)
			{
				long		i ;
				long		ac ;
				t_atom		*a ;
				t_atomarray *ptr1 = NULL ;
				t_atomarray *ptr2 = NULL ;
				
				ptr1 = linklist_getindex (x->resultat, 0) ;
				
				while (ptr1) 
					{
						long	ac ;
						t_atom	*a = NULL ;
						
						linklist_next (x->resultat, (void *)ptr1, (void *)&ptr2) ;
						
						atomarray_getatoms (ptr1, &ac, &a) ;
						atomarray_appendatoms (temp, ac, a) ;

						ptr1 = ptr2 ;	
					}
				
				linklist_clear (x->resultat) ;
				
				if (atomarray_getsize (temp) > argc)
					{
						long j ;
						
						for (j = atomarray_getsize (temp) ; j >= argc ; j--)
							{
								atomarray_chuckindex (temp, j) ;
							}
					}
				
				atomarray_getatoms (temp, &ac, &a) ;
				
				for (i = 0 ; i < ac ; i++)
					{
						long h = rand () % 100 ;
						
						if (h >= x->mutation)
							{
								if (atom_gettype (argv + i) == 1)
									{
										atom_setlong (a + i, atom_getlong (argv + i)) ;
									}
								else
									{
										atom_setlong (a + i, 0) ;
									}
							}
					}
													
				linklist_append (x->resultat, temp) ;
			}
		else 	
			{
				atomarray_appendatoms (temp, argc, argv) ;
				linklist_append (x->resultat, temp) ;
			}
						
		systhread_mutex_unlock (x->m_mutex) ;
		
		linklist_funall	(x->resultat, (method)outlet_linkfun, x) ;
		linklist_clear	(x->resultat) ;
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void outlet_linkfun (t_object *p, t_romeo *x)
	{
		t_atomarray *ptr = NULL ;
		
		if (ptr = (t_atomarray *)p)
			{
				long	ac ;
				t_atom	*a = NULL ;
				
				atomarray_getatoms (ptr, &ac, &a) ;
				
				outlet_list (x->m_out1, NULL, (short)ac, a) ;
			}
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void graf_effacer (t_romeo *x)
	{	
		long i ;
		
		x->bins[0 * TAILLE_VRAC] = 0 ;
		x->bins[TAILLE_DESCRIPTEURS * TAILLE_VRAC] = 1 ;

		for (i = 1 ; i < TAILLE_DESCRIPTEURS ; i++)
			{
				x->bins_topo[i] = 0 ;
			}
			
		x->bins_topo[0] = 1 ;
		x->bins_topo[TAILLE_DESCRIPTEURS] = 1 ;
		
		for (i = 0 ; i < TAILLE_VRAC ; i++)
			{
				long j ;
				
				if (i != 1)
					{
						x->graf_vrac[i].cardinal = 0 ;
					}
				
				for (j = 0 ; j < TAILLE_VRAC ; j++)
					{
						x->graf_vrac[i].enfants[j] = 0 ;
						x->graf_vrac[i].parents[j] = 0 ;
					}
			}
		
		x->graf_vrac[0].enfants[1] = 1 ;
		x->graf_vrac[1].parents[0] = 1 ;
		
		clear_fifo (x) ;
		
		for (i = 2 ; i < TAILLE_VRAC ; i++)
			{
				put_fifo (x, i) ;
			}
		
		x->nombre_concept = 0 ;
		
		x->graf_pos		= -1 ;
		x->graf_temp	= -1 ;
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void graf_ajouter (t_romeo *x) 
	{	
		long i ;
		
		for (i = 0 ; i < (TAILLE_DESCRIPTEURS + 1) ; i++)
			{
				x->temp_topo[i] = 0 ;
			}
		
		for (i = 0 ; i < (TAILLE_DESCRIPTEURS + 1) ; i++)
			{
				long j ;
				
				for (j = 0 ; j < x->bins_topo[i] ; j++)
					{
						long h ;
						long k ;
						
						h = x->bins[(i * TAILLE_VRAC) + j] ;
						k = concept_inclusion (x, x->graf_vrac[h].descripteurs) ;
												
						if (k == 2)
							{
								goto fin ;
							}
						else if (k == 1)
							{
								x->temp[(i * TAILLE_VRAC) + x->temp_topo[i]] = h ;
								x->temp_topo[i] ++ ;
							}
						else if (k == 0)
							{
								long t ;
								long g = 1 ;
								
								for (t = 0 ; t < x->temp_topo[x->cardinal_intersection] ; t++)
									{
										long p ;
										
										p = x->temp[(x->cardinal_intersection * TAILLE_VRAC) + t] ;
										
										if (intersection_identique (x, x->graf_vrac[p].descripteurs))
											{
												g = 0 ;
											}
									}
								
								if (g)
									{
										long n  ;
										
										if (n = nouveau_concept (x))
											{
												x->temp[(x->cardinal_intersection * TAILLE_VRAC) + x->temp_topo[x->cardinal_intersection]] = n ;
												x->temp_topo[x->cardinal_intersection] ++ ;
												
												x->graf_vrac[n].enfants[h] = 1 ;
												x->graf_vrac[h].parents[n] = 1 ;
												
												couverture_concept (x, h, n) ;
												
												if (intersection_identique (x, x->vecteur))
													{
														goto fin ;
													}
											}
									}
							}
					}
			}
		
	fin:	
		
		if (x->maj)
			{
				graf_cartographier (x) ;
				
				while (x->nombre_concept > (x->graf_max - 2))
					{
						graf_tuer (x) ;
						graf_cartographier (x) ;
					} 
				
				x->maj = 0 ;
			}
	}

void graf_cartographier (t_romeo *x)
	{	
		long i ;
		long k ;
		long j = 0 ;
		
		if (x->nombre_concept)
			{
				k = rand () % x->nombre_concept ;
			}
		else
			{
				k = -1 ;
				x->cible = -1 ;
			}
				
		x->bins[0 * TAILLE_VRAC] = 0 ;
		x->bins[TAILLE_DESCRIPTEURS * TAILLE_VRAC] = 1 ;

		for (i = 1 ; i < TAILLE_DESCRIPTEURS ; i++)
			{
				x->bins_topo[i] = 0 ;
			}
			
		x->bins_topo[0] = 1 ;
		x->bins_topo[TAILLE_DESCRIPTEURS] = 1 ;
		
		for (i = 2 ; i < TAILLE_VRAC ; i++)
			{
				long p ; 
				
				if (p = x->graf_vrac[i].cardinal)
					{
						x->bins[(p * TAILLE_VRAC) + x->bins_topo[p]] = i ;
						x->bins_topo[p] ++ ;
						
						if (j == k)
							{
								x->cible = i ;
							}
						
						j ++ ;
					}
			}
	}
	
void graf_tuer (t_romeo *x)
	{
		tuer_concept (x, x->cible) ;
		
		if (!x->nombre_concept)
			{
				x->graf_vrac[0].enfants[1] = 1 ;
				x->graf_vrac[1].parents[0] = 1 ;
			}
	}

long graf_naviguer (t_romeo *x)
	{
		long i, n ;
		long j = 0 ;				
		long k = 0 ;
		long urne[TAILLE_VRAC] ;
		
		if (x->graf_pos == -1)
			{
				x->graf_pos		= x->cible ;
				x->graf_temp	= x->cible ;
			}
		
		if (x->graf_pos > 1)
			{
				t_atomarray *ptr = NULL ;
				
				if (ptr = atomarray_new (0, NULL))
					{
						long t ;
						
						for (t = 0 ; t < TAILLE_DESCRIPTEURS ; t++)
							{
								if (get_itemset (x->graf_vrac[x->graf_pos].descripteurs, t))
									{
										t_atom a ;
										
										atom_setlong (&a, t) ;
										atomarray_appendatom (ptr, &a) ;
										
										k ++ ;
									}
							}
						
						linklist_append (x->resultat, ptr) ;
					}
			}
			
		for (i = 0 ; i < TAILLE_VRAC ; i++)
			{
				if (x->graf_vrac[x->graf_pos].parents[i] || x->graf_vrac[x->graf_pos].enfants[i])
					{
						urne[j] = i ;
						j ++ ;
					}
			}
		
		if (x->graf_pos < 2)
			{
				n = urne[rand () % j] ;
			}
		else
			{
				do {		
				n = urne[rand () % j] ;
				} while (n == x->graf_temp) ;
			}
		
		x->graf_temp = x->graf_pos ;
		x->graf_pos = n ;
		
		return k ;
	}

long graf_hasarder (t_romeo *x)
	{
		long i, h ;
		long k = 0 ;
		long j = 0 ;
		
		h = rand () % x->nombre_concept ;
		
		for (i = 2 ; i < TAILLE_VRAC ; i++)
			{
				if (x->graf_vrac[i].cardinal)
					{
						if (j == h)
							{
								t_atomarray *ptr = NULL ;
					
								if (ptr = atomarray_new (0, NULL))
									{
										long t ;
										
										for (t = 0 ; t < TAILLE_DESCRIPTEURS ; t++)
											{
												if (get_itemset (x->graf_vrac[i].descripteurs, t))
													{
														t_atom a ;
														
														atom_setlong (&a, t) ;
														atomarray_appendatom (ptr, &a) ;
													}
											}
										
										linklist_append (x->resultat, ptr) ;
									}
							
								k = 1 ;		
								break ;						
							}
						else
							{
								j ++ ;
							}
					}
			}
		
		return k ;
	}

long graf_lire (t_romeo *x, long n)
	{
		long i ;
		long k = 0 ;
		
		if (x->bins_topo[n])
			{
				for (i = 0 ; i < x->bins_topo[n] ; i++)
					{
						long		j ;
						t_atomarray *ptr = NULL ;
						
						j = x->bins[(n * TAILLE_VRAC) + i] ;
					
						if (ptr = atomarray_new (0, NULL))
							{
								long t ;
								
								for (t = 0 ; t < TAILLE_DESCRIPTEURS ; t++)
									{
										if (get_itemset (x->graf_vrac[j].descripteurs, t))
											{
												t_atom a ;
												
												atom_setlong (&a, t) ;
												atomarray_appendatom (ptr, &a) ;
											}
									}
								
								linklist_append (x->resultat, ptr) ;
							}
					}
				
				k = 1 ;		
			}
			
		return k ;
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

long nouveau_concept (t_romeo *x)
	{
		long i ;
		long k = 0 ;
		
		if (get_fifo (x))
			{
				k = x->val_fifo ;
				
				x->maj = 1 ;

				for (i = 0 ; i < TAILLE_BLOCS ; i++)
					{
						x->graf_vrac[k].descripteurs[i] = x->intersection[i] ;
					}

				x->graf_vrac[k].cardinal = x->cardinal_intersection ;
				
				x->nombre_concept ++ ;
			}
		
		return k ;
	}

void couverture_concept	(t_romeo *x, long h, long n)
	{
		long i ;
		
		for (i = 0 ; i < x->cardinal_intersection ; i++)
			{
				long j ;
				
				for (j = 0 ; j < x->temp_topo[i] ; j++)
					{
						long a = x->temp[(i * TAILLE_VRAC) + j] ;
						
						if (intersection_inclusion (x, x->graf_vrac[a].descripteurs))
							{
								long t ;
								long papa = 1 ;
																
								for (t = 0 ; t < TAILLE_VRAC ; t++)
									{
										if (x->graf_vrac[a].enfants[t])
											{
												if (intersection_inclusion (x, x->graf_vrac[t].descripteurs))
													{
														papa = 0 ;
														break ;
													}
											}
									}
								
								if (papa)
									{
										if (x->graf_vrac[h].parents[a])
											{
												x->graf_vrac[h].parents[a] = 0 ;
												x->graf_vrac[a].enfants[h] = 0 ;
											}
										
										x->graf_vrac[a].enfants[n] = 1 ;
										x->graf_vrac[n].parents[a] = 1 ;
									}
							}
					}
			}
	}

void tuer_concept (t_romeo *x, long n)
	{
		if (n > 1)
			{
				long i ;

				for (i = 0 ; i < TAILLE_VRAC ; i++)
					{
						if (x->graf_vrac[n].parents[i])
							{
								x->graf_vrac[i].enfants[n] = 0 ;
								x->graf_vrac[n].parents[i] = 0 ;
								
								if (sans_enfants (x, i))
									{
										tuer_concept (x, i) ;
									}
							}
						
						if (x->graf_vrac[n].enfants[i])
							{
								x->graf_vrac[i].parents[n] = 0 ;
								x->graf_vrac[n].enfants[i] = 0 ;
								
								if (sans_parents (x, i))
									{
										tuer_concept (x, i) ;
									}
							}
					}
				
				x->graf_vrac[n].cardinal = 0 ;
				
				x->nombre_concept -- ;
				put_fifo (x, n) ;
								
				if (x->graf_pos == n)
					{
						x->graf_pos		= -1 ;
						x->graf_temp	= -1 ;
					}
			}
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

long sans_parents (t_romeo *x, long a) 
	{
		long i ;
		long k = 1 ;
		
		for (i = 0 ; i < TAILLE_VRAC ; i++)
			{
				if (x->graf_vrac[a].parents[i])
					{
						k = 0 ;
						break ;
					}
			}
		
		return k ;
	}
	
long sans_enfants (t_romeo *x, long b) 
	{	
		long i ;
		long k = 1 ;
		
		for (i = 0 ; i < TAILLE_VRAC ; i++)
			{
				if (x->graf_vrac[b].enfants[i])
					{
						k = 0 ;
						break ;
					}
			}
		
		return k ;
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

long concept_inclusion (t_romeo *x, ulong *ptr)
	{
		long i ;
		long m = 0 ;
		long n = 0 ;
		long k = 2 ;
				
		for (i = 0 ; i < TAILLE_BLOCS ; i++)
			{
				x->intersection[i] = ptr[i] & x->vecteur[i] ; 

				if (ptr[i] != x->vecteur[i])
					{
						m = 1 ;
					}
					
				if (x->vecteur[i] != (x->vecteur[i] | ptr[i]))
					{
						n = 1 ;
					}
			}
		
		x->cardinal_intersection = size_itemset (x->intersection) ;
		
		if (n)
			{
				k = 0 ;
			}
		else if (m)
			{
				k = 1 ;
			}
			
		return k ;
	}

long intersection_identique (t_romeo *x, ulong *ptr)
	{	
		long i ;
		long k = 1 ;
		
		for (i = 0 ; i < TAILLE_BLOCS ; i++)
			{
				if (ptr[i] != x->intersection[i])
					{
						k = 0 ;
						break ;
					}
			}
			
		return k ;
	}

long intersection_inclusion (t_romeo *x, ulong *ptr) 
	{	
		long i ;
		long k = 1 ;
		
		for (i = 0 ; i < TAILLE_BLOCS ; i++)
			{
				if (x->intersection[i] != (x->intersection[i] | ptr[i]))
					{
						k = 0 ;
						break ;
					}
			}
			
		return k ;
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void set_itemset (ulong *ptr, long n) 
	{
		long i, p, m ;
		
		i = n / TAILLE_ULONG ;
		p = n % TAILLE_ULONG ;
		
		m = 1 << p ;
		
		*(ptr + i) |= m ;
	}
	
long get_itemset (ulong *ptr, long n) 
	{
		long	i, p ;
		ulong	temp ;
		
		i = n / TAILLE_ULONG ;
		p = n % TAILLE_ULONG ;
		
		temp = *(ptr + i) ;
	
		temp >>= p ;
		temp  &= 1 ;
		
		return temp != 0 ;
	}
	
long size_itemset (ulong *ptr) 
	{
		long i ;
		long k = 0 ;
		
		for (i = 0 ; i < TAILLE_BLOCS ; i++)
			{
				ulong n = *(ptr + i) ;
				
				while (n != 0)
					{
						k  += (n & 1) ;
						n >>= 1 ;
					}
			}
		
		return k ;
	}

// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

long put_fifo (t_romeo *x, long n) 
	{	
		long k = 0 ;
		
		if (x->fifo && ((x->queue_fifo + 1) != x->head_fifo) && !((x->queue_fifo == TAILLE_FIFO) && (x->head_fifo == 0)))
			{
				k = 1 ;
				
				x->fifo[x->queue_fifo] = n ;
				
				if (x->queue_fifo == TAILLE_FIFO)
					{
						x->queue_fifo = 0 ;
					}
				else
					{
						x->queue_fifo ++ ;
					}
			}
		
		return k ;
	}
	
long get_fifo (t_romeo *x)
	{
		long k = 0 ;
		
		if (x->fifo && (x->head_fifo != x->queue_fifo))
			{
				k = 1 ;
				
				x->val_fifo = x->fifo[x->head_fifo] ;
				
				if (x->head_fifo == TAILLE_FIFO)
					{
						x->head_fifo = 0 ;
					}
				else
					{
						x->head_fifo ++ ;
					}
			}
		
		return k ;
	}
	
void clear_fifo (t_romeo *x)
	{
		x->head_fifo	= 0 ;
		x->queue_fifo	= 0 ;
		x->val_fifo		= 0 ;
	}
	
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------

void romeo_dump (t_romeo *x, long n)
	{
		if (!x->lock_dump)
			{
				long ac ;
				long text_max = 50 ;
				char *affichage = NULL ;
				
				x->lock_dump = 1 ;
				
				post ("Nombre de concepts : %ld / %ld", x->nombre_concept, x->graf_max - 2) ;
				post ("---") ;

				if (affichage = (char *)sysmem_newptr (sizeof(char) * text_max))
					{
						long		i ;
						t_atom		*p = NULL ;
						t_atomarray *q = NULL ;
						
						if (q = atomarray_new (0, NULL))
							{
								for (i = 0 ; i < (TAILLE_DESCRIPTEURS + 1) ; i++)
									{
										long j ;
										
										for (j = 0 ; j < x->bins_topo[i] ; j++)
											{
												long t ;
												long c ;
												
												c = x->bins[(i * TAILLE_VRAC) + j] ;
												
												for (t = 0 ; t < TAILLE_DESCRIPTEURS ; t++)
													{
														if (get_itemset (x->graf_vrac[c].descripteurs, t))
															{
																t_atom a ;
																
																atom_setlong (&a, t) ;
																atomarray_appendatom (q, &a) ;
															}
													}
												
												atomarray_getatoms (q, &ac, &p) ;
												atom_gettext (ac, p, &text_max, &affichage, OBEX_UTIL_ATOM_GETTEXT_DEFAULT) ;
												
												if (ac == 0)
													{
														post ("0 : INF") ;
													}
												else if (ac == TAILLE_DESCRIPTEURS)
													{
														post ("1 : SUP") ;
													}
												else
													{
														post ("%ld : %s", c, affichage) ;
													}
												
												atomarray_clear (q) ;
												
												for (t = 0 ; t < TAILLE_VRAC ; t++)
													{
														if (x->graf_vrac[c].parents[t])
															{
																long z ;

																for (z = 0 ; z < TAILLE_DESCRIPTEURS ; z++)
																	{
																		if (get_itemset (x->graf_vrac[t].descripteurs, z))
																			{
																				t_atom a ;
																				
																				atom_setlong (&a, z) ;
																				atomarray_appendatom (q, &a) ;
																			}
																	}
																
																atomarray_getatoms (q, &ac, &p) ;
																atom_gettext (ac, p, &text_max, &affichage, OBEX_UTIL_ATOM_GETTEXT_DEFAULT) ;
																
																if (ac == 0)
																	{
																		post ("   parent : inf") ;
																	}
																else
																	{
																		post ("   parent : %s", affichage) ;
																	}
																
																atomarray_clear (q) ;
															}
													}
													
												for (t = 0 ; t < TAILLE_VRAC ; t++)
													{
														if (x->graf_vrac[c].enfants[t])
															{
																long z ;

																for (z = 0 ; z < TAILLE_DESCRIPTEURS ; z++)
																	{
																		if (get_itemset (x->graf_vrac[t].descripteurs, z))
																			{
																				t_atom a ;
																				
																				atom_setlong (&a, z) ;
																				atomarray_appendatom (q, &a) ;
																			}
																	}
																
																atomarray_getatoms (q, &ac, &p) ;
																atom_gettext (ac, p, &text_max, &affichage, OBEX_UTIL_ATOM_GETTEXT_DEFAULT) ;
																
																if (ac == TAILLE_DESCRIPTEURS)
																	{
																		post ("   enfant : sup") ;
																	}
																else
																	{
																		post ("   enfant : %s", affichage) ;
																	}
																
																atomarray_clear (q) ;
															}
													}
											}
									}
								
								object_free (q) ;
							}
				
						sysmem_freeptr (affichage) ;
					}
				
				clock_fdelay (x->m_clock, 1000.) ;
			}
	}
	
void romeo_task (t_romeo *x)
	{
		  x->lock_dump = 0 ;
	}
		
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -------------------------------------------------------------------------------------------------------------
// -----------------------------------------------------------------------------------------------------------:x