/*	

Zoulou ! 
							  
(Factor oracle with length repeated suffix).
						
merci : Matthieu Raffinot, Maxime Crochemore, et Cyril Allauzen ;
merci : Arnaud Lefebvre et Thierry Lecroq ;

Variation autour d'un "Breadth first search" pour "path".

Version / 19 décembre 2010.	
	
*/

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

#include "ext.h"							
#include "ext_obex.h"
#include "stdlib.h"
#include "math.h"	
#include "time.h"	
#include "ext_common.h"
#include "ext_systhread.h"	

// ----------------------------------------------------------------------------------------------------------------

#define MAX_RESULT			256
#define INIT_GRAF_SIZE		1024
#define TAILLE_GRAF_MAX		262144

#define FIFO_SIZE			256
#define BFS_MAX				512

#define DEF_SLICE			0
#define DEF_REPEAT			2
#define DEF_THRESHOLD		3
#define DEF_STRAIGHT		20
#define DEF_PATH_GO			1
#define DEF_PATH_BACKWARD	1

// ----------------------------------------------------------------------------------------------------------------

typedef struct _arc {
	long		dest ;
	long		val ;
	} t_arc ;
	
typedef struct _noeud {
	long		refer ;
	long		lrs ;
	long		straight_val ;
	long		path_state ;
	long		path_dist ;
	long		path_dad ;
	long		*path_val ;
	t_linklist	*list_arcs ;
	} t_noeud ;

// ----------------------------------------------------------------------------------------------------------------

typedef struct _zoulou {
	t_object			ob ;	
	t_noeud				**graf ;
	t_systhread_mutex   m_mutex ;
	t_atomarray			*slice_result ;
	t_atomarray			*path_result ;
	t_atomarray			*get_result ;
	long				*fifo ;
	long				graf_size ;             
	long				graf_pos ;			
	long				graf_index ;
	long				repeat ;
	long				straight ;
	long				threshold ;
	long				slice ;
	long				path_go ;
	long				path_backward ;
	long				temp_slice ;
	long				flag_slice ;
	long				flag_path ;
	long				lock_dump ;
	long				head_fifo ;
	long				queue_fifo ;
	long				val_fifo ;
	void				*m_clock ;
	void				*m_out1 ;
	void				*m_out2 ;
	void				*m_out3 ;
	void				*m_out4 ;
	} t_zoulou ;

// ----------------------------------------------------------------------------------------------------------------	

void	*zoulou_new			(t_symbol *s, long argc, t_atom *argv) ;
void	zoulou_bang			(t_zoulou *x) ;
void	zoulou_int			(t_zoulou *x, long n) ;
void	zoulou_clear		(t_zoulou *x) ;
void	zoulou_go			(t_zoulou *x, long n) ;
void	zoulou_path			(t_zoulou *x, long n) ;
void	zoulou_get			(t_zoulou *x) ;

void	graf_add			(t_zoulou *x, long n) ;

t_noeud	*noeud_new			(t_noeud *ptr) ;
t_arc	*arc_new			(long d, long v) ;
long	graf_exist			(t_noeud **p, long n, long v) ;       
long	arc_exist			(t_arc *aptr, long *v) ;
long	repeated_suffix		(t_noeud **p, long a, long b) ;  
long	common_suffix		(t_noeud **p, long a, long b) ;

long	graf_step			(t_zoulou *x) ;
void	graf_clear			(t_zoulou *x) ;

void	fill_slice_result	(t_zoulou *x, long ac, long av) ;

long	put_fifo			(t_zoulou *x, long n) ;
long	get_fifo			(t_zoulou *x) ;
void	path_search			(t_zoulou *x, long p, long n) ;

void	zoulou_dump			(t_zoulou *x, long n) ;
void	zoulou_task			(t_zoulou *x) ;
void	zoulou_assist		(t_zoulou *x, void *b, long m, long a, char *s) ;
void	zoulou_free			(t_zoulou *x) ;

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

static t_class	*zoulou_class ;

static t_symbol	*my_msg_straight	= NULL ;
static t_symbol	*my_msg_link		= NULL ;
static t_symbol	*my_msg_lrs			= NULL ;
static t_symbol	*my_msg_path		= NULL ;
static t_symbol	*my_msg_nopath		= NULL ;

int main (void)
	{	
		t_class *c = NULL ;
		
		my_msg_straight	= gensym ("straight") ;
		my_msg_link		= gensym ("arc") ;
		my_msg_lrs		= gensym ("lrs") ;
		my_msg_path		= gensym ("path") ;
		my_msg_nopath	= gensym ("no_path") ;
		
		c = class_new	("zoulou", (method)zoulou_new, (method)zoulou_free, (long)sizeof(t_zoulou), 0L, A_GIMME, 0) ;

		class_addmethod (c, (method)zoulou_assist,	"assist", A_CANT, 0) ; 
		class_addmethod (c, (method)zoulou_bang,	"bang", 0) ;
		class_addmethod (c, (method)zoulou_int,		"int", A_LONG, 0) ;
		class_addmethod (c, (method)zoulou_clear,	"clear", 0) ;
		class_addmethod (c, (method)zoulou_dump,	"dump_max", A_DEFLONG, 0) ;
		class_addmethod (c, (method)zoulou_go,		"go", A_DEFLONG, 0) ;
		class_addmethod (c, (method)zoulou_path,	"path", A_DEFLONG, 0) ;
		class_addmethod (c, (method)zoulou_get,		"get", 0) ;

		class_register	(CLASS_BOX, c) ; 
		
		CLASS_ATTR_LONG		(c, "lrs",				ATTR_FLAGS_NONE, t_zoulou, repeat) ;
		CLASS_ATTR_LONG		(c, "straight",			ATTR_FLAGS_NONE, t_zoulou, straight) ;
		CLASS_ATTR_LONG		(c, "threshold",		ATTR_FLAGS_NONE, t_zoulou, threshold) ;
		CLASS_ATTR_LONG		(c, "slice",			ATTR_FLAGS_NONE, t_zoulou, slice) ;
		CLASS_ATTR_LONG		(c, "path_go",			ATTR_FLAGS_NONE, t_zoulou, path_go) ;
		CLASS_ATTR_LONG		(c, "path_backward",	ATTR_FLAGS_NONE, t_zoulou, path_backward) ;
		
		CLASS_ATTR_LABEL	(c, "lrs",				ATTR_FLAGS_NONE, "Length Repeated Suffix Threshold") ;
		CLASS_ATTR_LABEL	(c, "straight",			ATTR_FLAGS_NONE, "Straight Navigate Percentage") ;
		CLASS_ATTR_LABEL	(c, "threshold",		ATTR_FLAGS_NONE, "Slice Threshold") ;
		CLASS_ATTR_LABEL	(c, "slice",			ATTR_FLAGS_NONE, "Slice Mode") ;
		CLASS_ATTR_LABEL	(c, "path_go",			ATTR_FLAGS_NONE, "Path Go Founded") ;
		CLASS_ATTR_LABEL	(c, "path_backward",	ATTR_FLAGS_NONE, "Path Jump Backward") ;
		
		CLASS_ATTR_FILTER_CLIP	(c, "lrs",				0, LONG_MAX) ;
		CLASS_ATTR_FILTER_CLIP	(c, "straight",			0, 100) ;
		CLASS_ATTR_FILTER_CLIP	(c, "threshold",		1, MAX_RESULT) ;
		CLASS_ATTR_FILTER_CLIP	(c, "slice",			0, 2) ;
		CLASS_ATTR_FILTER_CLIP	(c, "path_go",			0, 1) ;
		CLASS_ATTR_FILTER_CLIP	(c, "path_backward",	0, 1) ;
		
		CLASS_ATTR_ENUMINDEX	(c, "slice",			ATTR_FLAGS_NONE, "off maximum threshold") ;
		CLASS_ATTR_STYLE		(c, "path_go",			ATTR_FLAGS_NONE, "onoff") ;
		CLASS_ATTR_STYLE		(c, "path_backward",	ATTR_FLAGS_NONE, "onoff") ;
		
		CLASS_ATTR_ORDER (c, "lrs",				ATTR_FLAGS_NONE, "1") ;
		CLASS_ATTR_ORDER (c, "straight",		ATTR_FLAGS_NONE, "2") ;
		CLASS_ATTR_ORDER (c, "slice",			ATTR_FLAGS_NONE, "3") ;
		CLASS_ATTR_ORDER (c, "threshold",		ATTR_FLAGS_NONE, "4") ;
		CLASS_ATTR_ORDER (c, "path_go",			ATTR_FLAGS_NONE, "5") ;
		CLASS_ATTR_ORDER (c, "path_backward",	ATTR_FLAGS_NONE, "6") ;
		
		zoulou_class = c ;
		
		return MAX_ERR_NONE ;
	}

// ----------------------------------------------------------------------------------------------------------------

void *zoulou_new (t_symbol *s, long argc, t_atom *argv)
	{
		t_zoulou *x = NULL;
		
		if (x = (t_zoulou *)object_alloc (zoulou_class)) 
			{
				if (x->graf = (t_noeud **)sysmem_newptrclear (sizeof(t_noeud *) * INIT_GRAF_SIZE)) 
					{	
						t_noeud *ptr = NULL ;
						
						if (x->graf[0] = noeud_new (ptr))
							{
								srand (time(NULL)) ;
								
								x->graf_size		= INIT_GRAF_SIZE ;
								x->graf_index		= 1 ;
								x->graf_pos			= 0 ;
								x->temp_slice		= 0 ;
								x->flag_slice		= 0 ;
								x->flag_path		= 0 ;
								x->lock_dump		= 0 ;
								x->head_fifo		= 0 ;
								x->queue_fifo		= 0 ;
								x->val_fifo			= 0 ;
								x->repeat			= DEF_REPEAT ;
								x->straight			= DEF_STRAIGHT ;
								x->threshold		= DEF_THRESHOLD ;
								x->slice			= DEF_SLICE ;
								x->path_go			= DEF_PATH_GO ;
								x->path_backward	= DEF_PATH_BACKWARD ;
								
								x->fifo			= (long *)sysmem_newptr (sizeof(long) * FIFO_SIZE) ;
								
								x->slice_result	= atomarray_new (0, NULL) ;
								x->path_result	= atomarray_new (0, NULL) ;
								x->get_result	= atomarray_new (0, NULL) ;
								
								x->m_clock	= clock_new ((t_object *)x, (method)zoulou_task) ;
								
								x->m_out4	= outlet_new ((t_object *)x, NULL) ;
								x->m_out3	= listout ((t_object *)x) ;
								x->m_out2	= bangout ((t_object *)x) ;
								x->m_out1	= intout ((t_object *)x) ;
								
								systhread_mutex_new (&x->m_mutex, SYSTHREAD_MUTEX_NORMAL) ;

								attr_args_process (x, argc, argv) ;
							}
						else
							{
								object_free (x) ;
								x = NULL ;  
							}
					}
				else
					{
						object_free (x) ;
						x = NULL ;   
					}
			}
				
		return x ;
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void zoulou_bang (t_zoulou *x)
	{	
		if (x->graf_index < 2)
			{
				outlet_bang (x->m_out2) ;
			}
		else
			{
				long k ;
				
				systhread_mutex_lock (x->m_mutex) ;
						
				k = graf_step (x) ;
				
				systhread_mutex_unlock (x->m_mutex) ;
				
				outlet_int (x->m_out1, k) ;
			}
	}
	
void zoulou_int (t_zoulou *x, long n)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				if (x->graf_index < TAILLE_GRAF_MAX)
					{		
						graf_add (x, n) ;
					}
				
				systhread_mutex_unlock (x->m_mutex) ;
			}

		if (x->flag_slice)
			{
				t_atom	*av ;
				long	ac	;
				
				atomarray_getatoms (x->slice_result, &ac, &av) ;
				
				outlet_list (x->m_out3, NULL, (short)ac, av) ;
					
				x->flag_slice = 0 ;
			}
	}
	
void zoulou_clear (t_zoulou *x)
	{
		systhread_mutex_lock (x->m_mutex) ;
				
		graf_clear (x) ;
		
		systhread_mutex_unlock (x->m_mutex) ;
	}

void zoulou_go (t_zoulou *x, long n)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				if (n >= 0 && n < (x->graf_index - 1))
					{		
						x->graf_pos = n ;
					}
				
				systhread_mutex_unlock (x->m_mutex) ;
			}
	}

void zoulou_path (t_zoulou *x, long n)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				path_search (x, x->graf_pos, n) ;
				
				systhread_mutex_unlock (x->m_mutex) ;
			}
		
		if (x->flag_path)
			{
				t_atom	*av ;
				long	ac	;
				
				atomarray_getatoms (x->path_result, &ac, &av) ;
				
				outlet_anything (x->m_out4, my_msg_path, (short)ac, av) ;
					
				x->flag_path = 0 ;
			}
		else
			{
				outlet_anything (x->m_out4, my_msg_nopath, 0, NULL) ;
			}
	}
	
void zoulou_get (t_zoulou *x)           
	{	
		if (!systhread_mutex_trylock (x->m_mutex)) 
			{	
				if (x->get_result)
					{
						t_noeud *nptr = NULL ;
						t_arc	*aptr1 = NULL ;
						t_arc	*aptr2 = NULL ;
						t_atom	msg[3] ;
						
						nptr = x->graf[x->graf_pos] ;
						
						atom_setsym	 (msg, my_msg_lrs) ;
						atom_setlong (msg + 1, nptr->lrs) ;
						atom_setlong (msg + 2, nptr->refer) ;
						
						atomarray_appendatoms (x->get_result, 3, msg) ;
						
						aptr1 = linklist_getindex (nptr->list_arcs, 0) ;
						while (aptr1) 
							{
								linklist_next (nptr->list_arcs, (void *)aptr1, (void *)&aptr2) ;
								
								atom_setsym	 (msg, my_msg_link) ;
								atom_setlong (msg + 1, aptr1->val) ;
								atom_setlong (msg + 2, aptr1->dest) ;
						
								atomarray_appendatoms (x->get_result, 3, msg) ;
						
								aptr1 = aptr2 ;	
							} 
						
						atom_setsym	 (msg, my_msg_straight) ;
						atom_setlong (msg + 1, nptr->straight_val) ;
						atom_setlong (msg + 2, x->graf_pos + 1) ;
						
						atomarray_appendatoms (x->get_result, 3, msg) ;
					}
					
				systhread_mutex_unlock (x->m_mutex) ;
			}
			
		if (x->get_result)
			{
				t_atom	*av ;
				long	ac ;
						
				atomarray_getatoms (x->get_result, &ac, &av) ;
				
				if (ac) 
					{
						long	i ;
						t_atom	*ptr = NULL ;
						
						for (i = 0 ; i < ac ; i += 3)
							{
								ptr = av + i ;
								
								outlet_anything (x->m_out4, atom_getsym (ptr), (short)2, ptr + 1) ;
							}
						
						atomarray_clear (x->get_result) ;
					}
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void graf_add (t_zoulou *x, long n)
	{
		long	k, j, w ;
		t_noeud *nptr = NULL ;
		t_arc	*aptr = NULL ;
		long	index = x->graf_index ;
		
		if (index == x->graf_size)
			{
				t_noeud **temp = x->graf ;
			
				if (x->graf = (t_noeud **)sysmem_resizeptrclear (x->graf, sizeof(t_noeud *) * x->graf_size * 2))     
					{  
						x->graf_size = x->graf_size * 2 ;
					}
				else
					{
						x->graf = temp ; 
					}
			}
 
		if (index < x->graf_size)
			{
				t_noeud *ptr = NULL ;
				
				if (x->graf[index])
					{
						ptr = x->graf[index] ;
					}
				
				if (x->graf[index] = noeud_new (ptr)) 
					{
						nptr = x->graf[index - 1] ;
						nptr->straight_val = n ;
						
						j = nptr->refer ;
						k = index - 1 ;
						
						while (j > -1 && !graf_exist (x->graf, j, n))
							{
								nptr = x->graf[j] ;
								if (aptr = arc_new (index, n)) 
									{
										linklist_append (nptr->list_arcs, aptr) ;
									}
								k = j ;
								j = nptr->refer ;
							}
							
						if (j == -1)
							{
								w = 0 ;
							}
						else
							{
								w = graf_exist (x->graf, j, n) ;
							}
							
						nptr = x->graf[index] ;
						nptr->refer = w ;
						nptr->lrs = repeated_suffix (x->graf, k, nptr->refer) ;
						
						switch (x->slice) {
							case 0	:	break ;
							case 1	:	if (nptr->lrs < x->temp_slice && x->temp_slice >= x->threshold)
											{
												fill_slice_result (x, x->temp_slice, x->graf[index - 1]->refer) ;	
											}
										break ;
							case 2	:	if (nptr->lrs >= x->threshold)
											{
												fill_slice_result (x, x->threshold, x->graf[index]->refer) ;	
											}
										break ;	
							}
	
						x->temp_slice = nptr->lrs ;
						x->graf_index++ ;
					}
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
	
t_noeud *noeud_new (t_noeud *ptr)
	{
		t_noeud *nptr = NULL ;

		if (ptr)
			{
				object_free (ptr->list_arcs) ;
				sysmem_freeptr (ptr) ;
			}
				
		if (nptr = (t_noeud *)sysmem_newptrclear (sizeof(t_noeud))) 
			{
				nptr->refer			= -1 ;
				nptr->lrs			= 0 ;
				nptr->straight_val	= 0 ;
				nptr->path_state	= 0 ;
				nptr->path_dist		= -1 ;
				nptr->path_dad		= -1 ;
				nptr->path_val		= NULL ;
				nptr->list_arcs		= linklist_new () ;
				linklist_flags (nptr->list_arcs, OBJ_FLAG_MEMORY) ;
			}
			
		return nptr ;
	}

t_arc *arc_new (long d, long v)
	{
		t_arc *aptr = NULL ;
		
		if (aptr = (t_arc *)sysmem_newptrclear (sizeof(t_arc)))
			{
				aptr->dest = d ;
				aptr->val = v ;
			}
			
		return aptr ;
	}
	
long graf_exist (t_noeud **p, long n, long v)
	{
		long	k = 0 ;
		t_arc	*found = NULL ;
		t_noeud *nptr = NULL ;
		
		nptr = *(p + n) ;
		
		if (nptr->straight_val == v)
			{
				k = n + 1 ;
			}
		else
			{
				linklist_findfirst (nptr->list_arcs, (void *)&found, (t_cmpfn)arc_exist, &v) ;
				if (found)
					{
						k = found->dest ;
					}
			}
			
		return k ;
	}

long arc_exist (t_arc *aptr, long *v)	
	{
		return aptr->val == *v ;
	}

long repeated_suffix (t_noeud **p, long a, long b)
	{
		if (b == 0)
			{
				return 0 ;
			}
		else
			{
				return (common_suffix (p, a, b - 1)) + 1 ;
			}
	}

long common_suffix (t_noeud **p, long a, long b)
	{
		t_noeud *nptr1 = *(p + a) ;
		t_noeud *nptr2 = *(p + b) ;
		
		if (b == nptr1->refer)
			{
				return nptr1->lrs ;
			}
		else
			{
				while (nptr2->refer != nptr1->refer)
					{
						nptr2 = *(p + nptr2->refer) ;
					}
			}
			
		return fmin (nptr1->lrs, nptr2->lrs) ;
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
	
long graf_step (t_zoulou *x) 
	{
		long	*k = NULL ;
		t_noeud *nptr = NULL ;
		long	hasard = rand() % 100 ;
		
		if (x->graf_pos == ((x->graf_index) - 1)) 
			{
				x->graf_pos = 0 ;
			}
		
		if (hasard > x->straight)
			{
				nptr = x->graf[x->graf_pos] ;
			
				if (linklist_getsize(nptr->list_arcs))
					{
						t_arc *aptr = NULL ;
						
						linklist_shuffle (nptr->list_arcs) ;
						aptr = linklist_getindex (nptr->list_arcs, 0) ;
						k = &(aptr->val) ;
						x->graf_pos = aptr->dest ;
					}
			
				if (!k)
					{
						if (nptr->lrs > x->repeat)
							{
								x->graf_pos = nptr->refer ;
							}
					}
			}
		
		if (!k)
			{
				nptr = x->graf[x->graf_pos] ;		
				k = &(nptr->straight_val) ;
				x->graf_pos++ ;
			}
			
		if (k)
			{
				return *k ;
			}
		else
			{
				return -1 ;
			}
	}

void graf_clear (t_zoulou *x)
	{
		t_noeud *nptr = NULL ;
		
		nptr = x->graf[0] ;
		
		nptr->refer = -1 ;
		nptr->lrs = 0 ;
		nptr->straight_val = 0 ;
		linklist_clear (nptr->list_arcs) ;
		
		x->graf_index	= 1 ;
		x->graf_pos		= 0 ;
		x->temp_slice	= 0 ;
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void fill_slice_result (t_zoulou *x, long ac, long av)
	{
		if (x->slice_result && ac <= MAX_RESULT)
			{
				long	i ;
				t_noeud *nptr = NULL ;
				
				atomarray_clear (x->slice_result) ;
				
				for (i = 0 ; i < ac ; i++)
					{
						t_atom a ;
						
						nptr = x->graf[(av - ac + i)] ;
						
						atom_setlong (&a, nptr->straight_val) ;
						atomarray_appendatom (x->slice_result, &a) ;
					}
	
				x->flag_slice = 1 ;
			}
	}

// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

long put_fifo (t_zoulou *x, long n) 
	{
		long k = -1 ;
		
		if (x->fifo && ((x->queue_fifo + 1) != x->head_fifo) && !((x->queue_fifo == (FIFO_SIZE - 1)) && (x->head_fifo == 0)))
			{
				k = 0 ;
				
				x->fifo[x->queue_fifo] = n ;
				
				if (x->queue_fifo == (FIFO_SIZE - 1))
					{
						x->queue_fifo = 0 ;
					}
				else
					{
						x->queue_fifo ++ ;
					}
			}
		
		return k ;
	}
	
long get_fifo (t_zoulou *x)
	{
		long k = -1 ;
		
		if (x->fifo && (x->head_fifo != x->queue_fifo))
			{
				k = 0 ;
				
				x->val_fifo = x->fifo[x->head_fifo] ;
				
				if (x->head_fifo == (FIFO_SIZE - 1))
					{
						x->head_fifo = 0 ;
					}
				else
					{
						x->head_fifo ++ ;
					}
			}
		
		return k ;
	}

void path_search (t_zoulou *x, long p, long n)
	{
		if (p < x->graf_index)
			{
				long	i ;
				t_noeud	*nptr = NULL ;
				long	d = 1 ;
				long	del[BFS_MAX] ;
				
				nptr = x->graf[p] ;
				
				nptr->path_state	= 1 ;
				nptr->path_dist		= 0 ;
				
				del[0] = p ;
				
				if (!put_fifo (x, p))
					{
						long	error	= 0 ;
						long	block	= 0	;
						long	target	= -1 ;
						t_noeud *nptr1	= NULL ;
						t_noeud *nptr2	= NULL ;
						t_noeud *nptr3	= NULL ;
						t_noeud *nptr4	= NULL ;
						t_arc	*aptr1	= NULL ;
						t_arc	*aptr2	= NULL ;
						
						while (!error && target < 0 && !block) 
							{
								if (!(error = get_fifo (x)))
									{										
										nptr1 = x->graf[x->val_fifo] ;
										
										if (x->val_fifo < (x->graf_index - 1))
											{
												nptr2 = x->graf[x->val_fifo + 1] ;
																								
												if (!nptr2->path_state && !block && target < 0)
													{
														nptr2->path_state	= 1 ;
														nptr2->path_dist	= nptr1->path_dist + 1 ;
														nptr2->path_dad		= x->val_fifo ;
														nptr2->path_val		= &nptr1->straight_val ;
														
														del[d] = x->val_fifo + 1 ;
														d ++ ;
														
														if (d == BFS_MAX)
															{
																block = 1 ;
															}
														
														if ((x->val_fifo + 1) < x->graf_index)
															{
																error = put_fifo (x, (x->val_fifo + 1)) ;
															}
															
														if (nptr1->straight_val == n)
															{
																target = x->val_fifo + 1 ;
															}
													}
												
												if (nptr2->path_state && target < 0 && nptr1->straight_val == n)
													{
														nptr2->path_dist	= nptr1->path_dist + 1 ;
														nptr2->path_dad		= x->val_fifo ;
														nptr2->path_val		= &nptr1->straight_val ;
														
														target = x->val_fifo + 1 ;
													}
											}
											
										if (nptr1->lrs && x->path_backward)
											{
												nptr2 = x->graf[nptr1->refer] ;
												
												if (!nptr2->path_state && !block && target < 0)
													{
														nptr2->path_state	= 1 ;
														nptr2->path_dist	= nptr1->path_dist + 1 ;
														nptr2->path_dad		= x->val_fifo ;
														nptr2->path_val		= NULL ;
														
														del[d] = nptr1->refer ;
														d ++ ;
														
														if (d == BFS_MAX)
															{
																block = 1 ;
															}
														
														error = put_fifo (x, nptr1->refer) ;
													}
											}
										
										aptr1 = linklist_getindex (nptr1->list_arcs, 0) ;
										while (aptr1) 
											{
												linklist_next (nptr1->list_arcs, (void *)aptr1, (void *)&aptr2) ;	// @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
																							
												nptr2 = x->graf[aptr1->dest] ;
												
												if (!nptr2->path_state && !block && target < 0)
													{
														nptr2->path_state	= 1 ;
														nptr2->path_dist	= nptr1->path_dist + 1 ;
														nptr2->path_dad		= x->val_fifo ;
														nptr2->path_val		= &aptr1->val ;
														
														del[d] = aptr1->dest ;
														d ++ ;
														
														if (d == BFS_MAX)
															{
																block = 1 ;
															}
														
														error = put_fifo (x, aptr1->dest) ;
														
														if (aptr1->val == n)
															{
																target = aptr1->dest ;
															}
													}
												
												if (nptr2->path_state && target < 0 && aptr1->val == n)
													{
														nptr2->path_dist	= nptr1->path_dist + 1 ;
														nptr2->path_dad		= x->val_fifo ;
														nptr2->path_val		= &aptr1->val ;
														
														target = aptr1->dest ;
													}
												
												aptr1 = aptr2 ;	
											}	
										
										nptr1->path_state = 2 ;
									}
							}
						
						if (target >= 0)
							{	
								nptr3 = x->graf[target] ;
								
								if (x->path_result && nptr3->path_dist <= MAX_RESULT)
									{
										long	w = 0 ;
										long	j = 0 ;
										long	k = -1 ;
										long	temp[MAX_RESULT] ;
										
										while (nptr3->path_dad >= 0 && k != p)
											{
												k = nptr3->path_dad ;
												
												if (nptr3->path_val)
													{
														temp[w] = *(nptr3->path_val) ;
														w ++ ;
													}
												
												nptr3 = x->graf[k] ;
											}
										
										atomarray_clear (x->path_result) ;
										
										for (j = (w - 1) ; j >= 0 ; j--)
											{
												t_atom a ;
												
												atom_setlong (&a, temp[j]) ;
												atomarray_appendatom (x->path_result, &a) ;
											}
										
										x->flag_path = 1 ;
									}
								
								if (x->path_go)
									{
										x->graf_pos = target ;
									}
							}
						
						x->head_fifo	= 0 ;
						x->queue_fifo	= 0 ;
						
						for (i = 0 ; i < d ; i++)
							{
								nptr4 = x->graf[del[i]] ;
						
								nptr4->path_state	= 0 ;
								nptr4->path_dist	= -1 ;
								nptr4->path_dad		= -1 ;
								nptr4->path_val		= NULL ;
							}
					}
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------

void zoulou_dump (t_zoulou *x, long n)
	{
		if (!x->lock_dump)
			{
				long	i ;
				t_noeud *nptr = NULL ;
				t_arc	*aptr1 = NULL ;
				t_arc	*aptr2 = NULL ;
				
				x->lock_dump = 1 ;
				
				post ("Nombre de noeuds : %ld", x->graf_index) ;
				post ("---") ;
				
				for (i = MIN (n, x->graf_index) ; i < MIN (x->graf_index, n + 50) ; i++)
					{
						nptr = x->graf[i] ;
												
						post ("Noeud %ld : val %ld / ref %ld / lrs %ld", i, nptr->straight_val, nptr->refer, nptr->lrs) ;
							  
						aptr1 = linklist_getindex (nptr->list_arcs, 0) ;
						while (aptr1) 
							{
								linklist_next (nptr->list_arcs, (void *)aptr1, (void *)&aptr2) ;
								post ("          arc : val %ld / dest %ld", aptr1->val, aptr1->dest) ;
								aptr1 = aptr2 ;	
							}  
					} 
				
				clock_fdelay (x->m_clock, 1000.) ;
			}
	}
	
void zoulou_task (t_zoulou *x)
	{
		  x->lock_dump = 0 ;
	}

void zoulou_assist (t_zoulou *x, void *b, long m, long a, char *s)
	{
		if (m == ASSIST_INLET) 
			{ 
				sprintf (s, "(int) (bang) clear go get path") ;
			} 
		else 
			{	
				switch (a) {
					case 0 :
						sprintf (s, "Navigate") ;
						break ;
					case 1:
						sprintf (s, "Bang if empty") ;
						break ;
					case 2 :
						sprintf (s, "Repeated slices") ;
						break ;	
					case 3 :
						sprintf (s, "Get / path") ;
						break ;
				}
			}
	}

void zoulou_free (t_zoulou *x)
	{ 
		long i ;
		t_noeud *nptr = NULL ;
		
		for (i = 0 ; i < x->graf_size ; i++) 
			{
				if (nptr = x->graf[i])
					{
						object_free (nptr->list_arcs) ;
						sysmem_freeptr (nptr) ;
					}
			}
			
		if (x->graf)
			{
				sysmem_freeptr (x->graf) ;
			}
		
		if (x->fifo)
			{
				sysmem_freeptr (x->fifo) ;
			}
		
		if (x->slice_result)
			{
				object_free	(x->slice_result) ;
			}
		
		if (x->path_result)
			{
				object_free	(x->path_result) ;
			}
		
		if (x->get_result)
			{
				object_free	(x->get_result) ;
			}
		
		if (x->m_clock)
			{
				clock_unset (x->m_clock) ;
				object_free (x->m_clock) ;
			}
		
		if (x->m_mutex)
			{
				systhread_mutex_free (x->m_mutex) ;
			}
	}
	
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// ----------------------------------------------------------------------------------------------------------------
// --------------------------------------------------------------------------------------------------------------:x